@@ -24,9 +24,9 @@
 //! Users define a provider, with one or more _probe_ functions in the D language. For example:
 //!
 //! ```d
-//! provider test {
-//!     probe start(uint8_t);
-//!     probe stop(char*, uint8_t);
+//! provider my_provider {
+//!     probe start_work(uint8_t);
+//!     probe start_work(char*, uint8_t);
 //! };
 //! ```
 //!
@@ -42,11 +42,12 @@
 //!
 //! ```ignore
 //! #![feature(asm)]
+//! #![cfg_attr(target_os = "macos", feature(asm_sym))]
 //! usdt::dtrace_provider!("test.d");
 //! ```
 //!
 //! This procedural macro will generate a Rust macro for each probe defined in the provider. Note
-//! that including the `asm` feature is required; see [the notes](#notes) for a discussion. The
+//! that including the `asm` features are required; see [the notes](#notes) for a discussion. The
 //! `feature` directive and the invocation of `dtrace_provider` **should both be at the crate
 //! root**, i.e., `src/lib.rs` or `src/main.rs`.
 //!
@@ -54,10 +55,14 @@
 //!
 //! ```ignore
 //! let x: u8 = 0;
-//! test_start!(|| x);
+//! my_provider::start_work!(|| x);
 //! ```
 //!
-//! Note that `test_start!` is called with a closure which returns the arguments, rather than the
+//! We can see that the macros are defined in a module named by the provider, with one macro for
+//! each probe, with the same name. See [below](#configurable-names) for how this naming may be
+//! configured.
+//!
+//! Note that `start_work!` is called with a closure which returns the arguments, rather than the
 //! actual arguments themselves. See [below](#probe-arguments) for details. Additionally, as the
 //! probes are exposed as _macros_, they should be included in the crate root, before any other
 //! module or item which references them.
@@ -95,8 +100,8 @@
 //! #[usdt::provider]
 //! mod test {
 //!     use crate::Arg;
-//!     fn start(x: u8) {}
-//!     fn stop(arg: &Arg) {}
+//!     fn start_work(x: u8) {}
+//!     fn stop_work(arg: &Arg) {}
 //! }
 //! ```
 //!
@@ -105,7 +110,7 @@
 //! is used to access the named key of a JSON-encoded string. For example:
 //!
 //! ```bash
-//! $ dtrace -n 'stop { printf("%s", json(copyinstr(arg0), "ok.buffer[0]")); }'
+//! $ dtrace -n 'stop_work { printf("%s", json(copyinstr(arg0), "ok.buffer[0]")); }'
 //! ```
 //!
 //! would print the first element of the vector `Arg::buffer`.
@@ -122,6 +127,46 @@
 //!
 //! See the [probe_test_attr] example for a complete example implementing probes in Rust.
 //!
+//! ## Configurable names
+//!
+//! When using the attribute macro or build.rs versions of the code-generator, the names of the
+//! provider and/or probes may be configured. Specifically, the `probe_format` argument to the
+//! attribute macro or `Builder` method sets a format string used to generate the names of the
+//! probe macros. This can be any string, and will have the keys `{provider}` and `{probe}`
+//! interpolated to the actual names of the provider and probe. As an example, consider a provider
+//! named `foo` with a probe named `bar`, and a format string of `probe_{provider}_{probe}` -- the
+//! name of the generated probe macro will be `probe_foo_bar`.
+//!
+//! In addition, when using the attribute macro version, the name of the _provider_ as seen by
+//! DTrace can be configured. This defaults to the name of the provider module. For example,
+//! consider a module like this:
+//!
+//! ```ignore
+//! #[usdt::provider(provider = "foo")]
+//! mod probes {
+//!     fn bar() {
+//! }
+//! ```
+//!
+//! The probe `bar` will appear in DTrace as `foo:::bar`, and will be accessible in Rust via the
+//! macro `probes::bar!`. Note that it's not possible to rename the probe module when using the
+//! attribute macro version.
+//!
+//! Conversely, one can change the name of the generated provider _module_ when using the builder
+//! version, but not the name of the provider as it appears to DTrace. Given a file `"test.d"` that
+//! names a provider `foo` and a probe `bar`, consider this code:
+//!
+//! ```ignore
+//! usdt::Builder::new("test.d")
+//!     .module("probes")
+//!     .build()
+//!     .unwrap();
+//! ```
+//!
+//! This probe `bar` will appear in DTrace as `foo:::bar`, but will now be accessible in Rust via
+//! the macro `probes::bar!`. Note that it's not possible to rename the provider as it appears in
+//! DTrace when using the builder version.
+//!
 //! Examples
 //! --------
 //!
@@ -195,8 +240,8 @@
 //! purpose. It may be passed as any argument to a probe function, and is guaranteed to be unique
 //! between different invocations of the same probe. See the type's documentation for details.
 //!
-//! Notes
-//! -----
+//! Feature flags
+//! -------------
 //!
 //! The USDT crate relies on inline assembly to hook into DTrace. Unfortunatley this feature is
 //! unstable, and requires explicitly opting in with `#![feature(asm)]` as well as running with a
@@ -214,6 +259,35 @@
 //!
 //! The `asm` feature is a default of the `usdt` crate.
 //!
+//! ### Rust toolchain versions and the `asm_sym` flag
+//!
+//! The toolchain story is unfortunately more complicated than this. As of Rust 1.58.0-nightly
+//! (2021-10-29), the `asm` feature has been broken out into several more fine-grained features, to
+//! more quickly allow stabilization of the core inline assembly components. The result is that
+//! this crate requires the `asm_sym` feature on macOS target platforms.
+//!
+//! Unfortunately, because of the way the features were added (see [this pull
+//! request][asm-sym-feature-pr]), this version of Rust nightly is a Rubicon: the `usdt` crate, and
+//! crates using it, _cannot be built with compilers both before and after this version._
+//! Specifically, it's not possible to write the set of feature flags that would allow code to be
+//! compiled with a nightly toolchain before and after this version. If we _include_ the
+//! `feature(asm_sym)` directive with a toolchain of 1.57 or earlier, the compiler will generate an
+//! error because that feature isn't known for those versions. If we _omit_ the directive, it will
+//! compile with previous toolchains, but a newer one will generate an error because the feature
+//! flag is required for opting into the functionality used in the `usdt` crate's implementation on
+//! macOS.
+//!
+//! There's no great solution here. If you're developing an application, i.e., something that
+//! you're sure can be built with a specific toolchain such as with a `rust-toolchain` file, you
+//! can write the correct feature attribute for that toolchain version.
+//!
+//! If you're building a library, things are more complicated, because you don't know what
+//! toolchain a consuming application will choose to use. It's not possible to use a `build.rs`
+//! file or other code-generation mechanism, because inner attributes must generally be written
+//! directly at the top of the crate's root source file. A mechanism that _expands_ to the right
+//! tokens is not sufficient. The only real approach is to specify which versions of the toolchain
+//! are supported by your library in the documentation, as we've done here.
+//!
 //! Selecting the no-op implementation
 //! ----------------------------------
 //!
@@ -241,6 +315,9 @@
 //! [probe_test_build]: https://github.com/oxidecomputer/usdt/tree/master/probe-test-build
 //! [probe_test_attr]: https://github.com/oxidecomputer/usdt/tree/master/probe-test-attr
 //! [serde]: https://serde.rs
+//! [asm-features]: https://github.com/rust-lang/rust/pull/90348
+//! [asm-sym-feature-pr]: https://github.com/rust-lang/rust/pull/90348
+
 // Copyright 2021 Oxide Computer Company
 
 use std::path::{Path, PathBuf};
@@ -283,11 +360,18 @@ impl Builder {
         self
     }
 
-    /// Set the format for generated probe macros. The provided format may include
-    /// the tokens {provider} and {probe} which will be substituted with the names
-    /// of the provider and probe. The default format is "{provider}_{probe}"
-    pub fn format(mut self, format: &str) -> Self {
-        self.config.format = Some(format.to_string());
+    /// Set the format for the name of generated probe macros.
+    ///
+    /// The provided format may include the tokens `{provider}` and `{probe}`, which will be
+    /// substituted with the names of the provider and probe. The default is `"{probe}"`.
+    pub fn probe_format(mut self, format: &str) -> Self {
+        self.config.probe_format = Some(format.to_string());
+        self
+    }
+
+    /// Set the name of the module containing the generated probe macros.
+    pub fn module(mut self, module: &str) -> Self {
+        self.config.module = Some(module.to_string());
         self
     }
 
